// LWSDK sample for testing the HostDisplayInfo global.
// (This file must be build with an Objective-C++ compiler)
// This sample contains a single Layout generic plugin.

// This sample is meant for the Mac Cocoa-based LightWave beginning with 9.6.1
// It will demonstrate how a plugin can create a custom NSView and attach it
// to a host window such that it can receive mouse and keyboard input as well
// as draw itself.

// The host's display info structure provides access to a NSView-based object
// (referred to as a LWSDKView) that represents a container for interfacing with
// the host in terms of user interfacing.

// It is possible to further extend the host's LWSDKView by attaching new methods
// via Objective C/C++; however, most should find it adequate to use the LWSDKView
// as a parent container for their custom NSView-based object.

// It is essential to test the host product info for proper version at runtime
// because this Cocoa-based host is only valid for LW version 9.6.1 and greater.
// Previous versions return a Carbon-based WindowPtr.  Therefore, if the HostDisplayInfo
// is to be used and the plugin is to function with pre-9.6.1 and recent hosts,
// then a Carbon-based and a Cocoa-based implementation are both needed.

#import "DEMOView.h"

#include <lwdisplay.h>
#include <lwpanel.h>
#include <lwhost.h>
#include <lwgeneric.h>

@interface DEMOAlertDelegate: NSObject
{
}

- (void) alertDidEnd:(NSAlert *)alert returnCode:(int)returnCode contextInfo:(void *)contextInfo;
@end

@implementation DEMOAlertDelegate

- (void) alertDidEnd:(NSAlert *)alert returnCode:(int)returnCode contextInfo:(void *)contextInfo
{
}
@end

static void eventSheetModalRequest( LWControl *ctl, HostDisplayInfo *hostinfo)
{
    NSAlert *alert = [NSAlert alertWithMessageText:@"This requester is being presented by a plug-in."
        defaultButton:@"Alrighty Then" alternateButton:@"I Like Cheese"
        otherButton:nil informativeTextWithFormat: @"If you think this is interesting, click th"];
    [alert setAlertStyle:NSInformationalAlertStyle];
    DEMOAlertDelegate *alert_delegate = [[DEMOAlertDelegate alloc] init];
    [alert setDelegate:alert_delegate];
    [alert beginSheetModalForWindow:[hostinfo->host.view window] modalDelegate:alert_delegate didEndSelector:@selector(alertDidEnd:returnCode:contextInfo:) contextInfo:nil];
}

struct MyPanelData
{
    GlobalFunc *global;
    LWPanelID panel;
    NSView *overlay_view;
    DEMOView *demo_view;
    LWControl *lwcontrol_button_ask;
    
    MyPanelData( void)
    {
        global = 0;
        panel = 0;
        demo_view = 0;
        lwcontrol_button_ask = 0;
    }
};

static void handle_USERRESIZE( LWPanelID pan, void *user_data, int w, int h)
{
    MyPanelData *my_data = (MyPanelData*)user_data;

    [my_data->overlay_view setFrame:NSMakeRect( 0, 0, w, h)];
    NSRect overlay_rect = [my_data->overlay_view frame];

    NSRect demo_rect = overlay_rect;
    [my_data->demo_view setFrame:demo_rect];    

    LWValue ival = { LWT_INTEGER };
    int cx = CON_X( my_data->lwcontrol_button_ask);
    int cy = CON_Y( my_data->lwcontrol_button_ask);
    int cw = CON_W( my_data->lwcontrol_button_ask);
    int ch = CON_H( my_data->lwcontrol_button_ask);
    [my_data->demo_view setIgnoreFrame:NSMakeRect( cx, demo_rect.size.height - ch - cy, cw, ch)];
}

static void handle_USEROPEN( LWPanelID pan, void *user_data)
{
    MyPanelData *my_data = (MyPanelData*)user_data;

    LWPanelFuncs *pLWPanelFuncs = (LWPanelFuncs*)my_data->global( LWPANELFUNCS_GLOBAL, GFUSE_TRANSIENT);
    if( !pLWPanelFuncs)
        return;
    pLWPanelFuncs->globalFun = my_data->global;

    // the panel must be opened before we can attach our Cocoa-based GUI elements.
    LWValue ival = { LWT_INTEGER };
    int panel_width = PAN_GETW( pLWPanelFuncs, my_data->panel);
    int panel_height = PAN_GETH( pLWPanelFuncs, my_data->panel);

    // obtain host display info from panel.
    LWSDKView *parent_view = 0;
    pLWPanelFuncs->get( my_data->panel, PAN_HOSTDISPLAY, &parent_view);        
    if( parent_view)
    {
        // create cocoa-based elements and attach as children of the parent view (provided via PAN_HOSTDISPLAY)
        my_data->overlay_view = [[NSView alloc] initWithFrame:NSMakeRect( 0, 0, panel_width, panel_height)];
        [parent_view addSubview:my_data->overlay_view];

        NSRect rect = [my_data->overlay_view frame];

        my_data->demo_view = [[DEMOView alloc] initWithFrame:rect];
        [my_data->overlay_view addSubview:my_data->demo_view];
    }
}

static void handle_USERCLOSE( LWPanelID pan, void *user_data)
{
    MyPanelData *my_data = (MyPanelData*)user_data;

    [my_data->demo_view removeFromSuperviewWithoutNeedingDisplay];
    [my_data->demo_view release];
}

// Open/Close a modaless panel
XCALL_( static int )DEMO_HostDisplayInfo( int version, GlobalFunc *global, void *local, void *serverData)
{
    // local is LWLayoutGeneric*

    if ( version != LWLAYOUTGENERIC_VERSION )
        return AFUNC_BADVERSION;

    LWPanelFuncs *pLWPanelFuncs = (LWPanelFuncs*)global( LWPANELFUNCS_GLOBAL, GFUSE_TRANSIENT);
    if( !pLWPanelFuncs)
        return AFUNC_BADGLOBAL;

    pLWPanelFuncs->globalFun = global; 

    HostDisplayInfo *pMainHostDisplayInfo = (HostDisplayInfo*)global(LWHOSTDISPLAYINFO_GLOBAL, GFUSE_TRANSIENT);

    MyPanelData *my_data = new MyPanelData;
    my_data->global = global;
    
    my_data->panel = pLWPanelFuncs->create( "Cocoa Demonstration", NULL);
    if( !my_data->panel)
        return AFUNC_BADGLOBAL;

    LWValue ival = { LWT_INTEGER };
    LWPanControlDesc desc;

    pLWPanelFuncs->set( my_data->panel, PAN_USERDATA, (void*)my_data);
    pLWPanelFuncs->set( my_data->panel, PAN_USEROPEN, (void*)handle_USEROPEN);
    pLWPanelFuncs->set( my_data->panel, PAN_USERCLOSE, (void*)handle_USERCLOSE);
    pLWPanelFuncs->set( my_data->panel, PAN_USERRESIZE, (void*)handle_USERRESIZE);

    my_data->lwcontrol_button_ask = BUTTON_CTL( pLWPanelFuncs, my_data->panel, "Sheet Modal Request");
    CON_SETEVENT( my_data->lwcontrol_button_ask, eventSheetModalRequest, pMainHostDisplayInfo);
    int lw = CON_LW( my_data->lwcontrol_button_ask);
    ival.intv.value = 5 - lw;
    my_data->lwcontrol_button_ask->set( my_data->lwcontrol_button_ask, CTL_X, &ival );

    pLWPanelFuncs->open( my_data->panel, PANF_PASSALLKEYS | PANF_NOBUTT | PANF_RESIZE | PANF_BLOCKING);
    pLWPanelFuncs->destroy( my_data->panel);

    delete my_data;
    
    return AFUNC_OK;
}

extern "C" 
{
ServerRecord ServerDesc[] =
{
   { LWLAYOUTGENERIC_CLASS, "DEMO_HostDisplayInfo", DEMO_HostDisplayInfo },
   { NULL }
};
}
